/*
 * Copyright (c) 1995,1996
 * David M. Howard
 * daveh@hendrix.reno.nv.us
 * www.greatbasin.net/~daveh
 *
 * Permission to use, copy, modify, distribute and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  
 * David M. Howard makes no representations about the suitability of this
 * software for any purpose.  It is provided 'as is' without express or
 * implied warranty. Do not use it in systems where damage to property
 * or injury to people may occur in the event of a malfunction
 *
 * A supported version of this software including updates is available.
 * Consulting services for porting, modification, web user interface design
 * and implementation are available also.
 * Contact David M. Howard at the above email address for details.
 */
/*H*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "ews.h"
#include "ewsapp.h"
#include "ewspvt.h"

/* *** NEW *** FullRequest is substantially modifed */

/* ======================================================= */
/* HTTP DEFAULT REQUEST HANDLERS                           */
/* ======================================================= */
long em_FullRequest(em_workspace_t *wrk, em_session_t *sep)
{
   long result = S_404_NOTFND;

   // DISPATCH RULES
   // The dispatching of urls to the various handlers is based
   // on the order listed in this function. you can change the
   // default by reordering the calls to the various handler.
   //
   // the strncmp check to check a prepended path name to 
   // bin files is removed. Instead, all url searches start
   // at the virtual root, and the ordering of calls in this
   // function determine which is found first

   // NOTE That if you have files with the same names in both
   // the bin urls and the files urls, you won't be able
   // to get at the second one in the precedence order from
   // your browser.

   // the filesystem handler will default to a particular filename
   // that is specified as 'filedefault' in the initialization in
   // ewsw95t.c. That default only works at the top level, not in
   // subdirectories. for example, http://www.xxx.com/ will
   // default to http://www.xxx.com/index.html but
   // http://www.xxx.com/sub won't default to
   // http://www.xxx.com/sub/index.html
   // you can change the name of the default filename from 'index.html'
   // to whatever. IE 3.0 prefers 'default.htm' and Netscape likes
   // 'index.html' although you can change it on either.

   // 1. look for executable urls
   if (result == S_404_NOTFND) {
      result = app_BinRequest(&wrk->response,&wrk->request,&wrk->system,sep);
   }

   // 2. look for a filesystem url
   if (result == S_404_NOTFND) {
      result = app_FileRequest(&wrk->response,&wrk->request,&wrk->system,sep);
	  }

   // if it wasn't found, the result will be S_404_NOTFND at this point
   return result;
}

long em_SimpleRequest( em_workspace_t *wrk,em_session_t *sep)
{
   return em_FullRequest(wrk,sep);
}

void em_ApplicationInit(em_app_t *app)
{
   /* initialize application object with default functions */
   app->simple_request = em_SimpleRequest;
   app->full_request   = em_FullRequest;
}
