/*
* Copyright (c) 1995,1996
* David M. Howard
* daveh@hendrix.reno.nv.us
* www.greatbasin.net/~daveh
*
* Permission to use, copy, modify, distribute and sell this software
* and its documentation for any purpose is hereby granted without fee,
* provided that the above copyright notice appear in all copies and
* that both that copyright notice and this permission notice appear
* in supporting documentation.
* David M. Howard makes no representations about the suitability of this
* software for any purpose.  It is provided 'as is' without express or
* implied warranty.
*
* A supported version of this software including updates is available.
* Consulting services for porting, modification, web user interface design
* and implementation are available also.
* Contact David M. Howard at the above email address for details.
*/
/*H*/

#include <stdio.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include "ews.h"
#include "ewsapp.h"

// remove this define to link without stdio.h
#define SERVER_FILE_SUPPORT 1

#ifdef SERVER_FILE_SUPPORT

long app_FileBody(em_response_t *rsp,em_request_t *rqp,em_io_t *iop, em_session_t *sep)
{
   app_resp_t   *r;
   unsigned long  count;
   long           status;
   char          *bp;

   /* check input parameters */
   if ((rsp == 0)||(iop == 0)||(rqp ==0)||(sep == 0)) {
      /* internal server error */
      return S_500_SRVERR;
   }

   /* start with status OK */
   status = 0;

   /* get pointer to host specific structure */
   r = rsp->p; /* get pointer to internal structure */

   /* check it */
   if (r == 0) {
      /* internal server error */
      return S_500_SRVERR;
   }

   /* if no file open, quit with an error return code */
   if (r->file == 0) {
      return S_404_NOTFND;
   }

   /* write all bytes of the file to the io object */
   bp = r->buffer;

   /* NOTE feof needs the cast to FILE * because it is a macro */
   while(!feof((FILE *)r->file)) {
      /* read a buffer of data */
      count = fread(bp,1,RESPONSE_BUFFER_SIZE,r->file);
      if (count == 0) {
         /* end of file, quit */
         status = 0;
         break;
      }

      /* type mismatch between ulong and long could result in error     */
      /* if count is greater than max positive long value. check for it */
      if (count >= (unsigned long)LONG_MAX) {
         return S_500_SRVERR;
      }

      /* write it to the output function */
      status = iop->write(iop,sep->iod,bp,(long)count);
      if (status != 0) {
         /* an error writing to the output occurred */
         break;
      }
   }

   /* done with output, close the file */
   if (r->file != 0) {
      fclose(r->file);
   }

   return status;
}


long app_FileRequest(em_response_t *rsp,em_request_t *rqp,em_io_t *iop, em_session_t *sep)
{
   int             result;
   char            suffix[16];
   long            len;
   long            i;
   long            j;
   app_resp_t     *dresp = iop->p;
   struct _stat    statbuf;

   /* IF FOPEN FUNCTION IS 0, THERE IS NO FILE ACCESS */
   if (fopen == 0) {
      return S_404_NOTFND;
   }

   /* init the suffix string to empty */
   memset(suffix,0,sizeof(suffix));

   /* setup the request response */

   /* copy root file path */
   strcpy(dresp->filename,iop->fileroot);

   /* append file path and file name */
   if (strcmp(rqp->Url,"/") == 0) {
      /* got 'root' slash, default to index.htm */
      strcat(dresp->filename,iop->filedefault);
   }
   else {
      /* got something, build filename from url */
      strcat(dresp->filename,rqp->Url);
   }

   /* extract suffix from filename */
   len = (int)strlen(dresp->filename);

   /* backup in filename until a dot is found */
   for(i=len-1;i>0;i--) if (dresp->filename[i] == '.') break;

   /* if a dot is found, copy the suffix */
   if (i > 0) {
      /* found a dot, copy the suffix to the buffer */
      /* up to the length of the suffix             */
      j = 0;
      for(j=0;j < (long)sizeof(suffix) - 1;j++) {
         if (i < len) {
            suffix[j] = dresp->filename[i];
            i++;
         }
         else {
            suffix[j] = 0;
         }
      }
   }


   /* determine content length of requested file */
	/* this function is very system dependent     */
	/* if you don't have it, just assign 0 as the */
	/* content length and the only effect is that */
	/* the browser wont show the filesytsem during */
	/* page download */
	/* the following code is for windows 95       */

	result = _stat(dresp->filename,&statbuf);
	if (result == 0) {
	   // file is accessible, get the length
      rsp->content_length = statbuf.st_size;
   
      /* OK so far, try to open the file */
      dresp->file = fopen(dresp->filename,"rb");
      if (dresp->file == 0) {
         // file open failed, mark not found
         result = S_404_NOTFND;
      }
      else {
         // must get here, with filesize and file open
         // to continue ok
         result = S_200_OK;
      }
   }
   else {
      // stat failed, mark not found
      result = S_404_NOTFND;
   }

   /* at this point, if result is 0, a file is open */
   if (result == S_200_OK) {
      /* PROCESS THE FILE */

      /* determine content type from file suffix */
      if ((strcmp(".html",suffix) == 0)||(strcmp(".htm",suffix) == 0)) {
         rsp->content_type = em_GetHeaderString(EM_HDR_TEXT_HTML);
      }
      else if (strcmp(".gif",suffix) == 0) {
         rsp->content_type = em_GetHeaderString(EM_HDR_IMAGE_GIF);
      }
      else if (strcmp(".jpg",suffix) == 0) {
         rsp->content_type = em_GetHeaderString(EM_HDR_IMAGE_JPEG);
      }
      else if (strcmp(".class",suffix) == 0) {
         rsp->content_type = em_GetHeaderString(EM_HDR_TEXT_PLAIN);
      }
      else {
         /* unknown type, call it plain and let the browser worry about it */
         rsp->content_type = em_GetHeaderString(EM_HDR_TEXT_PLAIN);
      }

      /* load next buffer function */
      rsp->body = app_FileBody;

      /* static response data */
      rsp->p    = iop->p;

      result = S_200_OK;
   }
   else {
      /* can't open file or can't get status or is a directory */
      rsp->content_length = 0;
      rsp->content_type   = 0;
      rsp->body           = 0;
      rsp->p              = 0;
      result = S_404_NOTFND;
   }

   return result;
}

long app_FileHead(em_response_t *rsp,em_request_t *rqp,em_io_t *iop, em_session_t *sep)
{
   int result;
   app_resp_t *dresp = iop->p;

   /* IF FOPEN FUNCTION IS 0, THERE IS NO FILE ACCESS */
   if (fopen == 0) {
      return S_404_NOTFND;
   }

   /* setup the request response */

   /* copy root path */
   strcpy(dresp->filename,iop->fileroot);

   /* append file path and name */
   if (strcmp(rqp->Url,"/") == 0) {
      /* default to index.html */
      strcat(dresp->filename,iop->filedefault);
   }
   else {
      /* build filename from url */
      strcat(dresp->filename,rqp->Url);
   }

   /* start with indication of no file found and reset if file is found */
   rsp->content_length  = 0; /* length of response */
   rsp->body            = 0;
   rsp->p               = 0;

   /* make sure the file can be opened */
   dresp->file = fopen(dresp->filename,"rb");
   if (dresp->file == 0) {
      /* can't open file, mark it not found */
      result = S_404_NOTFND;
   }
   else {
      /* ok, file is found, not a directory and open */
      rsp->content_length  = 0;      /* length of response */
      rsp->body            = 0;      /* no body on head request */
      rsp->p               = &dresp; /* static response data */

      /* close file since this is a head request */
      fclose(dresp->file);

      result = S_200_OK;
   }

   /* mark file closed */
   dresp->file = 0;

   return result;
}
#else

// NO SERVER FILE SUPPORT, just return not found
long app_FileRequest(em_response_t *rsp,em_request_t *rqp,em_io_t *iop, em_session_t *sep)
{
return S_404_NOTFND;
}

#endif
